from typing import Dict

import dgl
import torch
from omegaconf import DictConfig
from torch import nn


class TreeLSTM(nn.Module):
    def __init__(self, config: DictConfig):
        super().__init__()
        self._encoder_size = config.encoder_size

        self._dropout = nn.Dropout(config.encoder_dropout)

        self._W_iou = nn.Linear(config.embedding_size, 3 * config.encoder_size)
        self._U_iou = nn.Linear(config.encoder_size, 3 * config.encoder_size, bias=False)

        self._W_f = nn.Linear(config.embedding_size, config.encoder_size)
        self._U_f = nn.Linear(config.encoder_size, config.encoder_size, bias=False)

        self._out_linear = nn.Linear(config.encoder_size, config.decoder_size)
        self._norm = nn.LayerNorm(config.decoder_size)
        self._tanh = nn.Tanh()

    def message_func(self, edges: dgl.udf.EdgeBatch) -> Dict:
        h_f = self._U_f(edges.src["h"])
        x_f = edges.dst["x_f"]
        f = torch.sigmoid(x_f + h_f)
        return {"Uh": self._U_iou(edges.src["h"]), "fc": edges.src["c"] * f}

    @staticmethod
    def reduce_func(nodes: dgl.udf.NodeBatch) -> Dict:
        return {"Uh_sum": torch.sum(nodes.mailbox["Uh"], dim=1), "fc_sum": torch.sum(nodes.mailbox["fc"], dim=1)}

    @staticmethod
    def apply_node_func(nodes: dgl.udf.NodeBatch) -> Dict:
        iou = nodes.data["x_iou"] + nodes.data["Uh_sum"]
        i, o, u = torch.chunk(iou, 3, 1)
        i, o, u = torch.sigmoid(i), torch.sigmoid(o), torch.tanh(u)

        c = i * u + nodes.data["fc_sum"]
        h = o * torch.tanh(c)

        return {"h": h, "c": c}

    def forward(self, graph: dgl.DGLGraph) -> torch.Tensor:
        x = self._dropout(graph.ndata["x"])

        # init matrices for message propagation
        number_of_nodes = graph.number_of_nodes()
        graph.ndata["x_iou"] = self._W_iou(x)
        graph.ndata["x_f"] = self._W_f(x)
        graph.ndata["h"] = graph.ndata["x"].new_zeros((number_of_nodes, self._encoder_size))
        graph.ndata["c"] = graph.ndata["x"].new_zeros((number_of_nodes, self._encoder_size))
        graph.ndata["Uh_sum"] = graph.ndata["x"].new_zeros((number_of_nodes, 3 * self._encoder_size))
        graph.ndata["fc_sum"] = graph.ndata["x"].new_zeros((number_of_nodes, self._encoder_size))

        # propagate nodes
        dgl.prop_nodes_topo(
            graph,
            message_func=self.message_func,
            reduce_func=self.reduce_func,
            apply_node_func=self.apply_node_func,
        )

        # [n nodes; encoder size]
        h = graph.ndata.pop("h")
        # [n nodes; decoder size]
        out = self._tanh(self._norm(self._out_linear(h)))
        return out



from argparse import ArgumentParser
from typing import cast

import torch
from commode_utils.common import print_config
from omegaconf import OmegaConf, DictConfig

from embeddings_for_trees.data.jsonl_data_module import JsonlTypedASTDatamodule
from embeddings_for_trees.models import TypedTreeLSTM2Seq
from embeddings_for_trees.utils.common import filter_warnings
from embeddings_for_trees.utils.test import test
from embeddings_for_trees.utils.train import train


def configure_arg_parser() -> ArgumentParser:
    arg_parser = ArgumentParser()
    arg_parser.add_argument("mode", help="Mode to run script", choices=["train", "test"])
    arg_parser.add_argument("-c", "--config", help="Path to YAML configuration file", type=str)
    return arg_parser


def train_typed_treelstm(config: DictConfig):
    filter_warnings()

    if config.print_config:
        print_config(config, fields=["model", "data", "train", "optimizer"])

    # Load data module
    data_module = JsonlTypedASTDatamodule(config.data, config.data_folder)
    data_module.prepare_data()
    data_module.setup()

    # Load model
    typed_treelstm2seq = TypedTreeLSTM2Seq(
        config.model, config.optimizer, data_module.vocabulary, config.train.teacher_forcing
    )

    train(typed_treelstm2seq, data_module, config)


def test_typed_treelstm(config: DictConfig):
    filter_warnings()

    # Load data module
    data_module = JsonlTypedASTDatamodule(config.data, config.data_folder)
    data_module.prepare_data()
    data_module.setup()

    # Load model
    typed_treelstm2seq = TypedTreeLSTM2Seq.load_from_checkpoint(config.checkpoint, map_location=torch.device("cpu"))

    test(typed_treelstm2seq, data_module, config.seed)


if __name__ == "__main__":
    __arg_parser = configure_arg_parser()
    __args = __arg_parser.parse_args()

    __config = cast(DictConfig, OmegaConf.load(__args.config))
    if __args.mode == "train":
        train_typed_treelstm(__config)
    else:
        test_typed_treelstm(__config)



import torch
from omegaconf import DictConfig

from embeddings_for_trees.models import TreeLSTM2Seq
from embeddings_for_trees.models.modules import TypedNodeEmbedding
from embeddings_for_trees.data.vocabulary import TypedVocabulary


class TypedTreeLSTM2Seq(TreeLSTM2Seq):
    def __init__(
        self,
        model_config: DictConfig,
        optimizer_config: DictConfig,
        vocabulary: TypedVocabulary,
        teacher_forcing: float = 0.0,
    ):
        super().__init__(model_config, optimizer_config, vocabulary, teacher_forcing)
        self._vocabulary: TypedVocabulary

    def _get_embedding(self) -> torch.nn.Module:
        return TypedNodeEmbedding(self._model_config, self._vocabulary)



import dgl
import torch
from omegaconf import DictConfig
from torch import nn

from embeddings_for_trees.utils.common import TOKEN, NODE, TYPE
from embeddings_for_trees.data.vocabulary import Vocabulary, TypedVocabulary


class NodeEmbedding(nn.Module):
    def __init__(self, config: DictConfig, vocabulary: Vocabulary):
        super().__init__()

        self._token_embedding = nn.Embedding(
            len(vocabulary.token_to_id), config.embedding_size, padding_idx=vocabulary.token_to_id[vocabulary.PAD]
        )
        self._node_embedding = nn.Embedding(
            len(vocabulary.node_to_id), config.embedding_size, padding_idx=vocabulary.node_to_id[vocabulary.PAD]
        )

    def forward(self, graph: dgl.DGLGraph) -> torch.Tensor:
        # [n nodes; embedding size]
        token_embedding = self._token_embedding(graph.ndata[TOKEN]).sum(1)
        # [n nodes; embedding size]
        node_embedding = self._node_embedding(graph.ndata[NODE])
        # [n nodes; 2 * embedding size]
        return token_embedding + node_embedding


class TypedNodeEmbedding(NodeEmbedding):
    def __init__(self, config: DictConfig, vocabulary: TypedVocabulary):
        super().__init__(config, vocabulary)
        self._type_embedding = nn.Embedding(
            len(vocabulary.type_to_id), config.embedding_size, padding_idx=vocabulary.type_to_id[vocabulary.PAD]
        )

    def forward(self, graph: dgl.DGLGraph) -> torch.Tensor:
        token_emb = self._token_embedding(graph.ndata[TOKEN]).sum(1)
        type_emb = self._type_embedding(graph.ndata[TYPE]).sum(1)
        node_emb = self._node_embedding(graph.ndata[NODE])
        return token_emb + type_emb + node_emb



from setuptools import setup, find_packages

VERSION = "1.0.0"

with open("README.md") as readme_file:
    readme = readme_file.read()

install_requires = [
    "torch>=1.9.0",
    "pytorch-lightning~=1.4.1",
    "torchmetrics~=0.4.0",
    "tqdm~=4.62.0",
    "wandb~=0.11.2",
    "omegaconf~=2.1.0",
    "commode-utils>=0.3.7",
    "dgl~=0.6.1",
]

setup_args = dict(
    name="embeddings-for-trees",
    version=VERSION,
    description="Set of PyTorch modules for developing and evaluating different algorithms for embedding trees.",
    long_description_content_type="text/markdown",
    long_description=readme,
    install_requires=install_requires,
    license="MIT",
    packages=find_packages(),
    author="Egor Spirin",
    author_email="spirin.egor@gmail.com",
    keywords=["treelstm", "tree-lstm", "embeddings-for-trees", "pytorch", "pytorch-lightning", "ml4code", "ml4se"],
    url="https://github.com/JetBrains-Research/embeddings-for-trees",
    download_url="https://pypi.org/project/embeddings-for-trees/",
)

if __name__ == "__main__":
    setup(**setup_args)



from typing import Tuple, List, Dict

import dgl
import torch
from commode_utils.losses import SequenceCrossEntropyLoss
from commode_utils.metrics import SequentialF1Score, ClassificationMetrics
from commode_utils.modules import LSTMDecoderStep, Decoder
from omegaconf import DictConfig
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities.types import EPOCH_OUTPUT
from torch.optim import Optimizer, AdamW
from torch.optim.lr_scheduler import _LRScheduler, LambdaLR
from torchmetrics import MetricCollection, Metric

from embeddings_for_trees.data.vocabulary import Vocabulary
from embeddings_for_trees.models.modules import NodeEmbedding, TreeLSTM


class TreeLSTM2Seq(LightningModule):
    def __init__(
        self,
        model_config: DictConfig,
        optimizer_config: DictConfig,
        vocabulary: Vocabulary,
        teacher_forcing: float = 0.0,
    ):
        super().__init__()
        self.save_hyperparameters()
        self._model_config = model_config
        self.__optim_config = optimizer_config
        self._vocabulary = vocabulary

        if vocabulary.SOS not in vocabulary.label_to_id:
            raise ValueError(f"Can't find SOS token in label to id vocabulary")

        self.__pad_idx = vocabulary.label_to_id[vocabulary.PAD]
        eos_idx = vocabulary.label_to_id[vocabulary.EOS]
        ignore_idx = [vocabulary.label_to_id[vocabulary.SOS]]
        metrics: Dict[str, Metric] = {
            f"{holdout}_f1": SequentialF1Score(pad_idx=self.__pad_idx, eos_idx=eos_idx, ignore_idx=ignore_idx)
            for holdout in ["train", "val", "test"]
        }
        self.__metrics = MetricCollection(metrics)

        self.__embedding = self._get_embedding()
        self.__encoder = TreeLSTM(model_config)
        decoder_step = LSTMDecoderStep(model_config, len(vocabulary.label_to_id), self.__pad_idx)
        self.__decoder = Decoder(
            decoder_step, len(vocabulary.label_to_id), vocabulary.label_to_id[vocabulary.SOS], teacher_forcing
        )

        self.__loss = SequenceCrossEntropyLoss(self.__pad_idx, reduction="batch-mean")

    @property
    def vocabulary(self) -> Vocabulary:
        return self._vocabulary

    def _get_embedding(self) -> torch.nn.Module:
        return NodeEmbedding(self._model_config, self._vocabulary)

    # ========== Main PyTorch-Lightning hooks ==========

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[_LRScheduler]]:
        optimizer = AdamW(
            self.parameters(),
            lr=self.__optim_config.lr,
            weight_decay=self.__optim_config.weight_decay,
        )

        def scheduler_lambda(epoch: int) -> float:
            return self.__optim_config.decay_gamma ** epoch

        scheduler = LambdaLR(optimizer, lr_lambda=scheduler_lambda)
        return [optimizer], [scheduler]

    def forward(  # type: ignore
        self,
        batched_trees: dgl.DGLGraph,
        output_length: int,
        target_sequence: torch.Tensor = None,
    ) -> torch.Tensor:
        batched_trees.ndata["x"] = self.__embedding(batched_trees)
        encoded_nodes = self.__encoder(batched_trees)
        output_logits = self.__decoder(encoded_nodes, batched_trees.batch_num_nodes(), output_length, target_sequence)
        return output_logits

    # ========== Model step ==========

    def _shared_step(self, batch: Tuple[torch.Tensor, dgl.DGLGraph], step: str) -> Dict:
        labels, graph = batch
        # [seq length; batch size; vocab size]
        logits = self(graph, labels.shape[0], labels) if step == "train" else self(graph, labels.shape[0])
        loss = self.__loss(logits[1:], labels[1:])

        with torch.no_grad():
            prediction = logits.argmax(-1)
            metric: ClassificationMetrics = self.__metrics[f"{step}_f1"](prediction, labels)

        return {
            f"{step}/loss": loss,
            f"{step}/f1": metric.f1_score,
            f"{step}/precision": metric.precision,
            f"{step}/recall": metric.recall,
        }

    def training_step(self, batch: Tuple[torch.Tensor, dgl.DGLGraph], batch_idx: int) -> Dict:  # type: ignore
        result = self._shared_step(batch, "train")
        self.log_dict(result, on_step=True, on_epoch=False)
        self.log("f1", result["train/f1"], prog_bar=True, logger=False)
        return result["train/loss"]

    def validation_step(self, batch: Tuple[torch.Tensor, dgl.DGLGraph], batch_idx: int) -> Dict:  # type: ignore
        result = self._shared_step(batch, "val")
        return result["val/loss"]

    def test_step(self, batch: Tuple[torch.Tensor, dgl.DGLGraph], batch_idx: int) -> Dict:  # type: ignore
        result = self._shared_step(batch, "test")
        return result["test/loss"]

    # ========== On epoch end ==========

    def _shared_epoch_end(self, step_outputs: EPOCH_OUTPUT, step: str):
        with torch.no_grad():
            losses = [so if isinstance(so, torch.Tensor) else so["loss"] for so in step_outputs]
            mean_loss = torch.stack(losses).mean()
            metric = self.__metrics[f"{step}_f1"].compute()
            log = {
                f"{step}/loss": mean_loss,
                f"{step}/f1": metric.f1_score,
                f"{step}/precision": metric.precision,
                f"{step}/recall": metric.recall,
            }
        self.log_dict(log, on_step=False, on_epoch=True)

    def training_epoch_end(self, step_outputs: EPOCH_OUTPUT):
        self._shared_epoch_end(step_outputs, "train")

    def validation_epoch_end(self, step_outputs: EPOCH_OUTPUT):
        self._shared_epoch_end(step_outputs, "val")

    def test_epoch_end(self, step_outputs: EPOCH_OUTPUT):
        self._shared_epoch_end(step_outputs, "test")
